#!/usr/bin/env python

from plistlib import loads, dumps, FMT_BINARY
from pathlib import Path
import pprint
from xattr import xattr
import os
import os.path
import sys
import tempfile
import subprocess
import datetime
from . import _applescript

# this was inspired by osx-tags by "Ben S / scooby" and is published under
# the same MIT license. See: https://github.com/scooby/osx-tags

# TODO: What to do about colors
# TODO: Add ability to remove key instead of just clear contents
# TODO: check what happens if OSXMetaData.__init__ called with invalid file--should result in error but saw one case where it didn't


# what to import
__all__ = ["OSXMetaData"]

# color labels
_COLORNAMES = {
    "None": 0,
    "Gray": 1,
    "Green": 2,
    "Purple": 3,
    "Blue": 4,
    "Yellow": 5,
    "Red": 6,
    "Orange": 7,
}

_COLORIDS = {
    0: "None",
    1: "Gray",
    2: "Green",
    3: "Purple",
    4: "Blue",
    5: "Yellow",
    6: "Red",
    7: "Orange",
}

_VALID_COLORIDS = "01234567"
_MAX_FINDERCOMMENT = 750  # determined through trial & error with Finder
_MAX_WHEREFROM = (
    1024
)  # just picked something....todo: need to figure out what max length is

_TAGS = "com.apple.metadata:_kMDItemUserTags"
_FINDER_COMMENT = "com.apple.metadata:kMDItemFinderComment"
_WHERE_FROM = "com.apple.metadata:kMDItemWhereFroms"
_DOWNLOAD_DATE = "com.apple.metadata:kMDItemDownloadedDate"


class _NullsInString(Exception):
    """Nulls in string."""


def _onError(e):
    sys.stderr.write(str(e) + "\n")


class _Tags:
    def __init__(self, xa: xattr):
        self.__attrs = xa
        self.__load_tags()

    def __tag_split(self, tag):
        """
        Extracts the color information from a Finder tag.
        """
        parts = tag.rsplit("\n", 1)
        if len(parts) == 1:
            return parts[0], 0
        elif (
            len(parts[1]) != 1 or parts[1] not in _VALID_COLORIDS
        ):  # Not a color number
            return tag, 0
        else:
            return parts[0], int(parts[1])

    def __load_tags(self):
        self.__tags = {}
        try:
            self.__tagvalues = self.__attrs[_TAGS]
            # load the binary plist value
            self.__tagvalues = loads(self.__tagvalues)
            for x in self.__tagvalues:
                (tag, color) = self.__tag_split(x)
                self.__tags[tag] = color
                # self.__tags = [self.__tag_strip_color(x) for x in self.__tagvalues]
        except KeyError:
            self.__tags = None
        if self.__tags:
            self.__tag_set = set(self.__tags.keys())
        else:
            self.__tag_set = set([])

    def __iter__(self):
        self.__load_tags()
        self.__tag_list = list(self.__tag_set)
        self.__tag_count = len(self.__tag_list)
        self.__tag_counter = 0
        return self

    def __next__(self):
        if self.__tag_counter < self.__tag_count:
            tag = self.__tag_list[self.__tag_counter]
            self.__tag_counter += 1
            return tag
        else:
            raise StopIteration

    def __len__(self):
        self.__load_tags()
        return len(self.__tag_set)

    def __repr__(self):
        self.__load_tags()
        return repr(self.__tag_set)

    def __str__(self):
        self.__load_tags()
        return ", ".join(self.__tag_set)

    def add(self, tag):
        if not isinstance(tag, str):
            raise TypeError("Tags must be strings")
        self.__load_tags()
        tags = set(map(self.__tag_normalize, self.__tag_set))
        tags.add(self.__tag_normalize(tag))
        self.__write_tags(*tags)

    def update(self, *tags):
        if not all(isinstance(tag, str) for tag in tags):
            raise TypeError("Tags must be strings")
        self.__load_tags()
        old_tags = set(map(self.__tag_normalize, self.__tag_set))
        new_tags = old_tags.union(set(map(self.__tag_normalize, tags)))
        self.__write_tags(*new_tags)

    def clear(self):
        try:
            self.__attrs.remove(_TAGS)
        except (IOError, OSError):
            pass

    def remove(self, tag):
        self.__load_tags()
        if not isinstance(tag, str):
            raise TypeError("Tags must be strings")
        tags = set(map(self.__tag_normalize, self.__tag_set))
        tags.remove(self.__tag_normalize(tag))
        self.__write_tags(*tags)

    def discard(self, tag):
        self.__load_tags()
        if not isinstance(tag, str):
            raise TypeError("Tags must be strings")
        tags = set(map(self.__tag_normalize, self.__tag_set))
        tags.discard(self.__tag_normalize(tag))
        self.__write_tags(*tags)

    def __iadd__(self, tag):
        self.add(tag)
        return self

    def __write_tags(self, *tags):
        """
        Overwrites the existing tags with the iterable of tags provided.
        """
        if not all(isinstance(tag, str) for tag in tags):
            raise TypeError("Tags must be strings")
        tag_plist = dumps(list(map(self.__tag_normalize, tags)), fmt=FMT_BINARY)
        self.__attrs.set(_TAGS, tag_plist)

    def __tag_colored(self, tag, color):
        """
        Sets the color of a tag.

        Parameters:
        tag(str): a tag name
        color(int): an integer from 1 through 7

        Return:
        (str) the tag with encoded color.
        """
        return "{}\n{}".format(self.__tag_nocolor(tag), color)

    def __tag_normalize(self, tag):
        """
        Ensures a color is set if not none.
        :param tag: a possibly non-normal tag.
        :return: A colorized tag.
        """
        tag, color = self.__tag_split(tag)
        if tag.title() in _COLORNAMES:
            # ignore the color passed and set proper color name
            return self.__tag_colored(tag.title(), _COLORNAMES[tag.title()])
        else:
            return self.__tag_colored(tag, color)

    def __tag_nocolor(self, tag):
        """
        Removes the color information from a Finder tag.
        """
        return tag.rsplit("\n", 1)[0]


class OSXMetaData:
    def __init__(self, fname):

        self.__fname = Path(fname)
        if not os.path.exists(self.__fname):
            raise (ValueError("file does not exist: ", fname))

        try:
            self.__attrs = xattr(self.__fname)
        except (IOError, OSError) as e:
            quit(_onError(e))

        # setup applescript for writing finder comments
        self.__scpt_set_finder_comment = _applescript.AppleScript(
            """
            on run {fname, fc}
	            set theFile to fname
	            set theComment to fc
	            tell application "Finder" to set comment of (POSIX file theFile as alias) to theComment
            end run
            """
        )

        # self.__setfc_script_fh = tempfile.NamedTemporaryFile(suffix=".scpt",prefix="osxmd",mode="w",delete=False)
        # self.__setfc_script = self.__setfc_script_fh.name
        # fd = self.__setfc_script_fh.file
        # fd.write(_APPLESCRIPT_SET_FINDER_COMMENT)
        # fd.close()
        # print("applescript = %s" % self.__setfc_script)

        # initialize meta data
        self.__tags = {}
        self.__findercomment = None
        self.__wherefrom = []
        self.__downloaddate = None

        self.tags = _Tags(self.__attrs)

        # TODO: Lot's of repetitive code here
        # need to read these dynamically
        """ Get Finder comment """
        self.__load_findercomment()

        """ Get Where From (for downloaded files) """
        self.__load_download_wherefrom()

        """ Get Download Date (for downloaded files) """
        self.__load_download_date()

    # @property
    # def colors(self):
    #     """ return list of color labels from tags
    #         do not return None (e.g. ignore tags with no color)
    #     """
    #     colors = []
    #     if self.__tags:
    #         for t in self.__tags.keys():
    #             c = self.__tags[t]
    #             if c == 0: continue
    #             colors.append(_COLORIDS[c])
    #         return colors
    #     else:
    #         return None

    def __del__(self):
        pass
        # print("removing temp file: %s" % self.__setfc_script)
        # os.remove(self.__setfc_script)

    def __load_findercomment(self):
        try:
            self.__fcvalue = self.__attrs[_FINDER_COMMENT]
            # load the binary plist value
            self.__findercomment = loads(self.__fcvalue)
        except KeyError:
            self.__findercomment = None

    @property
    def finder_comment(self):
        self.__load_findercomment()
        return self.__findercomment

    @finder_comment.setter
    def finder_comment(self, fc):
        """
        TODO: this creates a temporary script file which gets runs by osascript every time
              not very efficient.  Perhaps use py-applescript in the future but that increases
              dependencies + PyObjC
        """
        if fc is None:
            fc = ""
        elif not isinstance(fc, str):
            raise TypeError("Finder comment must be strings")

        if len(fc) > _MAX_FINDERCOMMENT:
            raise ValueError(
                "Finder comment limited to %d characters" % _MAX_FINDERCOMMENT
            )

        fname = self.__fname.resolve().as_posix()

        self.__scpt_set_finder_comment.run(fname, fc)

        self.__load_findercomment()

    def __load_download_wherefrom(self):
        try:
            self.__wfvalue = self.__attrs[_WHERE_FROM]
            # load the binary plist value
            self.__wherefrom = loads(self.__wfvalue)
        except KeyError:
            self.__wherefrom = None

    @property
    def where_from(self):
        self.__load_download_wherefrom()
        return self.__wherefrom

    @where_from.setter
    def where_from(self, wf):
        if wf is None:
            wf = []
        elif not isinstance(wf, list):
            raise TypeError("Where from must be a list of one or more URL strings")

        for w in wf:
            if len(w) > _MAX_WHEREFROM:
                raise ValueError(
                    "Where from URL limited to %d characters" % _MAX_WHEREFROM
                )

        wf_plist = dumps(wf, fmt=FMT_BINARY)
        self.__attrs.set(_WHERE_FROM, wf_plist)
        self.__load_download_wherefrom()

    def __load_download_date(self):
        try:
            self.__ddvalue = self.__attrs[_DOWNLOAD_DATE]
            # load the binary plist value
            # returns an array with a single datetime.datetime object
            self.__downloaddate = loads(self.__ddvalue)[0]
        except KeyError:
            self.__downloaddate = None

    @property
    def download_date(self):
        self.__load_download_date()
        return self.__downloaddate

    @download_date.setter
    def download_date(self, dt):
        if dt is None:
            dt = []
        elif not isinstance(dt, datetime.datetime):
            raise TypeError("Download date must be a datetime object")

        dt_plist = dumps([dt], fmt=FMT_BINARY)
        self.__attrs.set(_DOWNLOAD_DATE, dt_plist)
        self.__load_download_date()

    @property
    def name(self):
        return self.__fname.resolve().as_posix()



#!/usr/bin/env python

import osxmetadata
from osxmetadata import *

print(osxmetadata.__file__)

md = OSXMetaData("/Users/rhet/Dropbox (Personal)/Code/osxmetadata/osxmetadata/t.py")
md.tags.update("test", "foo")
print("tags = ", md.tags)
print(md.tags)
for tag in md.tags:
    print("got a tag: ", tag)

print(md.tags)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# setup.py script for osxmetdata
#
# Copyright (c) 2018 Rhet Turnbull, rturnbull+git@gmail.com
# All rights reserved.
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:

# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# from distutils.core import setup
from setuptools import setup

# read the contents of README file
from os import path

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="osxmetadata",
    version="0.95.6",
    description="Read and write meta data, such as tags/keywords, on Mac OS X files",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Rhet Turnbull",
    author_email="rturnbull+git@gmail.com",
    url="https://github.com/RhetTbull/",
    project_urls={"GitHub": "https://github.com/RhetTbull/osxmetadata"},
    download_url="https://github.com/RhetTbull/osxmetadata",
    packages=["osxmetadata"],
    license="License :: OSI Approved :: MIT License",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: MacOS X",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: MacOS :: MacOS X",
        "Programming Language :: Python",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    install_requires=["xattr", "pyobjc"],
)



""" aecodecs -- Convert from common Python types to Apple Event Manager types and vice-versa. """

import datetime, struct, sys

from Foundation import NSAppleEventDescriptor, NSURL

from . import kae


__all__ = ["Codecs", "AEType", "AEEnum"]


######################################################################


def fourcharcode(code):
    """ Convert four-char code for use in NSAppleEventDescriptor methods.
		
		code : bytes -- four-char code, e.g. b'utxt'
		Result : int -- OSType, e.g. 1970567284
	"""
    return struct.unpack(">I", code)[0]


#######


class Codecs:
    """ Implements mappings for common Python types with direct AppleScript equivalents. Used by AppleScript class. """

    kMacEpoch = datetime.datetime(1904, 1, 1)
    kUSRF = fourcharcode(kae.keyASUserRecordFields)

    def __init__(self):
        # Clients may add/remove/replace encoder and decoder items:
        self.encoders = {
            NSAppleEventDescriptor.class__(): self.packdesc,
            type(None): self.packnone,
            bool: self.packbool,
            int: self.packint,
            float: self.packfloat,
            bytes: self.packbytes,
            str: self.packstr,
            list: self.packlist,
            tuple: self.packlist,
            dict: self.packdict,
            datetime.datetime: self.packdatetime,
            AEType: self.packtype,
            AEEnum: self.packenum,
        }
        if sys.version_info.major < 3:  # 2.7 compatibility
            self.encoders[unicode] = self.packstr

        self.decoders = {
            fourcharcode(k): v
            for k, v in {
                kae.typeNull: self.unpacknull,
                kae.typeBoolean: self.unpackboolean,
                kae.typeFalse: self.unpackboolean,
                kae.typeTrue: self.unpackboolean,
                kae.typeSInt32: self.unpacksint32,
                kae.typeIEEE64BitFloatingPoint: self.unpackfloat64,
                kae.typeUTF8Text: self.unpackunicodetext,
                kae.typeUTF16ExternalRepresentation: self.unpackunicodetext,
                kae.typeUnicodeText: self.unpackunicodetext,
                kae.typeLongDateTime: self.unpacklongdatetime,
                kae.typeAEList: self.unpackaelist,
                kae.typeAERecord: self.unpackaerecord,
                kae.typeAlias: self.unpackfile,
                kae.typeFSS: self.unpackfile,
                kae.typeFSRef: self.unpackfile,
                kae.typeFileURL: self.unpackfile,
                kae.typeType: self.unpacktype,
                kae.typeEnumeration: self.unpackenumeration,
            }.items()
        }

    def pack(self, data):
        """Pack Python data.
			data : anything -- a Python value
			Result : NSAppleEventDescriptor -- an AE descriptor, or error if no encoder exists for this type of data
		"""
        try:
            return self.encoders[data.__class__](data)  # quick lookup by type/class
        except (KeyError, AttributeError) as e:
            for (
                type,
                encoder,
            ) in (
                self.encoders.items()
            ):  # slower but more thorough lookup that can handle subtypes/subclasses
                if isinstance(data, type):
                    return encoder(data)
        raise TypeError(
            "Can't pack data into an AEDesc (unsupported type): {!r}".format(data)
        )

    def unpack(self, desc):
        """Unpack an Apple event descriptor.
			desc : NSAppleEventDescriptor
			Result : anything -- a Python value, or the original NSAppleEventDescriptor if no decoder is found
		"""
        decoder = self.decoders.get(desc.descriptorType())
        # unpack known type
        if decoder:
            return decoder(desc)
            # if it's a record-like desc, unpack as dict with an extra AEType(b'pcls') key containing the desc type
        rec = desc.coerceToDescriptorType_(fourcharcode(kae.typeAERecord))
        if rec:
            rec = self.unpackaerecord(rec)
            rec[AEType(kae.pClass)] = AEType(struct.pack(">I", desc.descriptorType()))
            return rec
            # return as-is
        return desc

        ##

    def _packbytes(self, desctype, data):
        return NSAppleEventDescriptor.descriptorWithDescriptorType_bytes_length_(
            fourcharcode(desctype), data, len(data)
        )

    def packdesc(self, val):
        return val

    def packnone(self, val):
        return NSAppleEventDescriptor.nullDescriptor()

    def packbool(self, val):
        return NSAppleEventDescriptor.descriptorWithBoolean_(int(val))

    def packint(self, val):
        if (-2 ** 31) <= val < (2 ** 31):
            return NSAppleEventDescriptor.descriptorWithInt32_(val)
        else:
            return self.pack(float(val))

    def packfloat(self, val):
        return self._packbytes(kae.typeFloat, struct.pack("d", val))

    def packbytes(self, val):
        return self._packbytes(kae.typeData, val)

    def packstr(self, val):
        return NSAppleEventDescriptor.descriptorWithString_(val)

    def packdatetime(self, val):
        delta = val - self.kMacEpoch
        sec = delta.days * 3600 * 24 + delta.seconds
        return self._packbytes(kae.typeLongDateTime, struct.pack("q", sec))

    def packlist(self, val):
        lst = NSAppleEventDescriptor.listDescriptor()
        for item in val:
            lst.insertDescriptor_atIndex_(self.pack(item), 0)
        return lst

    def packdict(self, val):
        record = NSAppleEventDescriptor.recordDescriptor()
        usrf = desctype = None
        for key, value in val.items():
            if isinstance(key, AEType):
                if key.code == kae.pClass and isinstance(
                    value, AEType
                ):  # AS packs records that contain a 'class' property by coercing the packed record to the descriptor type specified by the property's value (assuming it's an AEType)
                    desctype = value
                else:
                    record.setDescriptor_forKeyword_(
                        self.pack(value), fourcharcode(key.code)
                    )
            else:
                if not usrf:
                    usrf = NSAppleEventDescriptor.listDescriptor()
                usrf.insertDescriptor_atIndex_(self.pack(key), 0)
                usrf.insertDescriptor_atIndex_(self.pack(value), 0)
        if usrf:
            record.setDescriptor_forKeyword_(usrf, self.kUSRF)
        if desctype:
            newrecord = record.coerceToDescriptorType_(fourcharcode(desctype.code))
            if newrecord:
                record = newrecord
            else:  # coercion failed for some reason, so pack as normal key-value pair
                record.setDescriptor_forKeyword_(
                    self.pack(desctype), fourcharcode(key.code)
                )
        return record

    def packtype(self, val):
        return NSAppleEventDescriptor.descriptorWithTypeCode_(fourcharcode(val.code))

    def packenum(self, val):
        return NSAppleEventDescriptor.descriptorWithEnumCode_(fourcharcode(val.code))

        #######

    def unpacknull(self, desc):
        return None

    def unpackboolean(self, desc):
        return desc.booleanValue()

    def unpacksint32(self, desc):
        return desc.int32Value()

    def unpackfloat64(self, desc):
        return struct.unpack("d", bytes(desc.data()))[0]

    def unpackunicodetext(self, desc):
        return desc.stringValue()

    def unpacklongdatetime(self, desc):
        return self.kMacEpoch + datetime.timedelta(
            seconds=struct.unpack("q", bytes(desc.data()))[0]
        )

    def unpackaelist(self, desc):
        return [
            self.unpack(desc.descriptorAtIndex_(i + 1))
            for i in range(desc.numberOfItems())
        ]

    def unpackaerecord(self, desc):
        dct = {}
        for i in range(desc.numberOfItems()):
            key = desc.keywordForDescriptorAtIndex_(i + 1)
            value = desc.descriptorForKeyword_(key)
            if key == self.kUSRF:
                lst = self.unpackaelist(value)
                for i in range(0, len(lst), 2):
                    dct[lst[i]] = lst[i + 1]
            else:
                dct[AEType(struct.pack(">I", key))] = self.unpack(value)
        return dct

    def unpacktype(self, desc):
        return AEType(struct.pack(">I", desc.typeCodeValue()))

    def unpackenumeration(self, desc):
        return AEEnum(struct.pack(">I", desc.enumCodeValue()))

    def unpackfile(self, desc):
        url = bytes(
            desc.coerceToDescriptorType_(fourcharcode(kae.typeFileURL)).data()
        ).decode("utf8")
        return NSURL.URLWithString_(url).path()


#######


class AETypeBase:
    """ Base class for AEType and AEEnum.
	
		Notes:
		
		- Hashable and comparable, so may be used as keys in dictionaries that map to AE records.
	"""

    def __init__(self, code):
        """
			code : bytes -- four-char code, e.g. b'utxt'
		"""
        if not isinstance(code, bytes):
            raise TypeError("invalid code (not a bytes object): {!r}".format(code))
        elif len(code) != 4:
            raise ValueError("invalid code (not four bytes long): {!r}".format(code))
        self._code = code

    code = property(
        lambda self: self._code, doc="bytes -- four-char code, e.g. b'utxt'"
    )

    def __hash__(self):
        return hash(self._code)

    def __eq__(self, val):
        return val.__class__ == self.__class__ and val.code == self._code

    def __ne__(self, val):
        return not self == val

    def __repr__(self):
        return "{}({!r})".format(self.__class__.__name__, self._code)


##


class AEType(AETypeBase):
    """An AE type. Maps to an AppleScript type class, e.g. AEType(b'utxt') <=> 'unicode text'."""


class AEEnum(AETypeBase):
    """An AE enumeration. Maps to an AppleScript constant, e.g. AEEnum(b'yes ') <=> 'yes'."""
